@@ -1,23 +1,24 @@
-/* -*- c-file-style: "java"; indent-tabs-mode: nil; fill-column: 78 -*- 
+/* -*- c-file-style: "java"; indent-tabs-mode: nil; tab-width: 4; fill-column: 78 -*-
  *
  * distcc -- A simple distributed compiler system
  *
  * Copyright (C) 2002, 2003, 2004 by Martin Pool <mbp@samba.org>
+ * Copyright 2007 Google Inc.
  *
  * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
+ * modify it under the terms of the GNU General Public License
+ * as published by the Free Software Foundation; either version 2
+ * of the License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
  *
- * This program is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * General Public License for more details.
- * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
- * USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
+ * USA.
  */
 
 
@@ -71,7 +72,7 @@
  **/
 
 
-#include "config.h"
+#include <config.h>
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -79,6 +80,7 @@
 #include <string.h>
 #include <fcntl.h>
 #include <errno.h>
+#include <assert.h>
 
 #include <sys/stat.h>
 
@@ -89,7 +91,7 @@
 #include "snprintf.h"
 
 
-static int dcc_argv_append(char *argv[], char *toadd)
+int dcc_argv_append(char **argv, char *toadd)
 {
     int l = dcc_argv_len(argv);
     argv[l] = toadd;
@@ -103,7 +105,7 @@ static void dcc_note_compiled(const char *input_file, const char *output_file)
 
     input_base = dcc_find_basename(input_file);
     output_base = dcc_find_basename(output_file);
-        
+
     rs_log(RS_LOG_INFO|RS_LOG_NONAME,
            "compile from %s to %s", input_base, output_base);
 }
@@ -118,6 +120,11 @@ static void dcc_note_compiled(const char *input_file, const char *output_file)
  * This code is called on both the client and the server, though they use the
  * results differently.
  *
+ * This function makes a copy of the arguments, modified to ensure that
+ * the arguments include '-o <filename>'.  This is returned in *ret_newargv.
+ * The copy is dynamically allocated and the caller is responsible for
+ * deallocating it.
+ *
  * @returns 0 if it's ok to distribute this compilation, or an error code.
  **/
 int dcc_scan_args(char *argv[], char **input_file, char **output_file,
@@ -158,10 +165,13 @@ int dcc_scan_args(char *argv[], char **input_file, char **output_file,
             } else if (!strcmp(a, "-MG") || !strcmp(a, "-MP")) {
                 /* These just modify the behaviour of other -M* options and do
                  * nothing by themselves. */
-            } else if (!strcmp(a, "-MF") || !strcmp(a, "-MT") || 
+            } else if (!strcmp(a, "-MF") || !strcmp(a, "-MT") ||
                        !strcmp(a, "-MQ")) {
-                /* as above but with extra argument */
+                /* As above but with extra argument. */
                 i++;
+            } else if (!strncmp(a, "-MF", 3) || !strncmp(a, "-MT", 3) ||
+                       !strncmp(a, "-MQ", 3)) {
+                /* As above, without extra argument. */
             } else if (a[1] == 'M') {
                 /* -M(anything else) causes the preprocessor to
                     produce a list of make-style dependencies on
@@ -171,6 +181,14 @@ int dcc_scan_args(char *argv[], char **input_file, char **output_file,
                     to distribute it even if we could. */
                 rs_trace("%s implies -E (maybe) and must be local", a);
                 return EXIT_DISTCC_FAILED;
+            } else if (!strcmp(a, "-march=native")) {
+                rs_trace("-march=native generates code for local machine; "
+                         "must be local");
+                return EXIT_DISTCC_FAILED;
+            } else if (!strcmp(a, "-mtune=native")) {
+                rs_trace("-mtune=native optimizes for local machine; "
+                         "must be local");
+                return EXIT_DISTCC_FAILED;
             } else if (str_startswith("-Wa,", a)) {
                 /* Look for assembler options that would produce output
                  * files and must be local.
@@ -196,6 +214,10 @@ int dcc_scan_args(char *argv[], char **input_file, char **output_file,
             } else if (str_startswith("-x", a)) {
                 rs_log_info("gcc's -x handling is complex; running locally");
                 return EXIT_DISTCC_FAILED;
+            } else if (str_startswith("-dr", a)) {
+                rs_log_info("gcc's debug option %s may write extra files; "
+                            "running locally", a);
+                return EXIT_DISTCC_FAILED;
             } else if (!strcmp(a, "-c")) {
                 seen_opt_c = 1;
             } else if (!strcmp(a, "-o")) {
@@ -246,7 +268,7 @@ int dcc_scan_args(char *argv[], char **input_file, char **output_file,
     if (!*output_file) {
         /* This is a commandline like "gcc -c hello.c".  They want
          * hello.o, but they don't say so.  For example, the Ethereal
-         * makefile does this. 
+         * makefile does this.
          *
          * Note: this doesn't handle a.out, the other implied
          * filename, but that doesn't matter because it would already
@@ -258,7 +280,7 @@ int dcc_scan_args(char *argv[], char **input_file, char **output_file,
          * preprocessing" rather than "stop after compilation." */
         if (seen_opt_s) {
             if (dcc_output_from_source(*input_file, ".s", &ofile))
-                return EXIT_DISTCC_FAILED; 
+                return EXIT_DISTCC_FAILED;
         } else if (seen_opt_c) {
             if (dcc_output_from_source(*input_file, ".o", &ofile))
                 return EXIT_DISTCC_FAILED;
@@ -296,8 +318,8 @@ int dcc_scan_args(char *argv[], char **input_file, char **output_file,
 int dcc_set_action_opt(char **a, const char *new_c)
 {
     int gotone = 0;
-    
-    for (; *a; a++) 
+
+    for (; *a; a++)
         if (!strcmp(*a, "-c") || !strcmp(*a, "-S")) {
             *a = strdup(new_c);
             if (*a == NULL) {
@@ -321,6 +343,7 @@ int dcc_set_action_opt(char **a, const char *new_c)
 
 /**
  * Change object file or suffix of -o to @p ofname
+ * Frees the old value, if it exists.
  *
  * It's crucially important that in every case where an output file is
  * detected by dcc_scan_args(), it's also correctly identified here.
@@ -329,22 +352,26 @@ int dcc_set_action_opt(char **a, const char *new_c)
 int dcc_set_output(char **a, char *ofname)
 {
     int i;
-    
-    for (i = 0; a[i]; i++) 
-        if (0 == strcmp(a[i], "-o")  &&  a[i+1] != NULL) {
-            rs_trace("changed output from \"%s\" to \"%s\"", a[i+1],
-                     ofname);
-            a[i+1] = ofname;
+
+    for (i = 0; a[i]; i++)
+        if (0 == strcmp(a[i], "-o") && a[i+1] != NULL) {
+            rs_trace("changed output from \"%s\" to \"%s\"", a[i+1], ofname);
+            free(a[i+1]);
+            a[i+1] = strdup(ofname);
+            if (a[i+1] == NULL) {
+                rs_log_crit("failed to allocate space for output parameter");
+                return EXIT_OUT_OF_MEMORY;
+            }
             dcc_trace_argv("command after", a);
             return 0;
         } else if (0 == strncmp(a[i], "-o", 2)) {
             char *newptr;
-
+            rs_trace("changed output from \"%s\" to \"%s\"", a[i]+2, ofname);
+            free(a[i]);
             if (asprintf(&newptr, "-o%s", ofname) == -1) {
                 rs_log_crit("failed to allocate space for output parameter");
                 return EXIT_OUT_OF_MEMORY;
             }
-            
             a[i] = newptr;
             dcc_trace_argv("command after", a);
             return 0;
@@ -354,9 +381,9 @@ int dcc_set_output(char **a, char *ofname)
     return EXIT_DISTCC_FAILED;
 }
 
-
 /**
- * Change input file to @p ifname; called on compiler.
+ * Change input file to a copy of @p ifname; called on compiler.
+ * Frees the old value.
  *
  * @todo Unify this with dcc_scan_args
  *
@@ -370,12 +397,125 @@ int dcc_set_input(char **a, char *ifname)
     for (i =0; a[i]; i++)
         if (dcc_is_source(a[i])) {
             rs_trace("changed input from \"%s\" to \"%s\"", a[i], ifname);
-            a[i] = ifname;
+            free(a[i]);
+            a[i] = strdup(ifname);
+            if (a[i] == NULL) {
+                rs_log_crit("failed to allocate space for input parameter");
+                return EXIT_OUT_OF_MEMORY;
+            }
             dcc_trace_argv("command after", a);
             return 0;
         }
-    
+
     rs_log_error("failed to find input file");
     return EXIT_DISTCC_FAILED;
 }
 
+/* Subroutine of dcc_expand_preprocessor_options().
+ * Calculate how many extra arguments we'll need to convert
+ * a "-Wp,..." option into regular gcc options.
+ * Returns the number of extra arguments needed.
+ */
+static int count_extra_args(char *dash_Wp_option) {
+    int extra_args = 0;
+    char *comma = dash_Wp_option + strlen("-Wp");
+    while (comma != NULL) {
+        char *opt = comma + 1;
+        comma = strchr(opt, ',');
+        if (str_startswith("-MD,", opt) ||
+            str_startswith("-MMD,", opt))
+        {
+            char *filename = comma + 1;
+            comma = strchr(filename, ',');
+            extra_args += 3;  /* "-MD", "-MF", filename. */
+        } else {
+            extra_args++;
+        }
+    }
+    return extra_args;
+}
+
+/* Subroutine of dcc_expand_preprocessor_options().
+ * Convert a "-Wp,..." option into one or more regular gcc options.
+ * Copy the resulting gcc options to dest_argv, which should be
+ * pre-allocated by the caller.
+ * Destructively modifies dash_Wp_option as it goes.
+ * Returns 0 on success, nonzero for error (out of memory).
+ */
+static int copy_extra_args(char **dest_argv, char *dash_Wp_option,
+                           int extra_args) {
+    int i = 0;
+    char *comma = dash_Wp_option + strlen("-Wp");
+    while (comma != NULL) {
+        char *opt = comma + 1;
+        comma = strchr(opt, ',');
+        if (comma) *comma = '\0';
+        dest_argv[i] = strdup(opt);
+        if (!dest_argv[i]) return EXIT_OUT_OF_MEMORY;
+        i++;
+        if (strcmp(opt, "-MD") == 0 || strcmp(opt, "-MMD")) {
+            char *filename;
+            if (!comma) {
+                rs_log_warning("'-Wp,-MD' or '-Wp,-MMD' option is missing "
+                               "filename argument");
+                break;
+            }
+            filename = comma + 1;
+            comma = strchr(filename, ',');
+            if (comma) *comma = '\0';
+            dest_argv[i] = strdup("-MF");
+            if (!dest_argv[i]) return EXIT_OUT_OF_MEMORY;
+            i++;
+            dest_argv[i] = strdup(filename);
+            if (!dest_argv[i]) return EXIT_OUT_OF_MEMORY;
+            i++;
+        }
+    }
+    assert(i == extra_args);
+    return 0;
+}
+
+
+/*
+ * Convert any "-Wp," options into regular gcc options.
+ * We do this because it simplifies the command-line
+ * option handling elsewhere; this is the only place
+ * that needs to parse "-Wp," options.
+ * Returns 0 on success, nonzero for error (out of memory).
+ *
+ * The argv array pointed to by argv_ptr when this function
+ * is called must have been dynamically allocated.  It remains
+ * the caller's responsibility to deallocate it.
+ */
+int dcc_expand_preprocessor_options(char ***argv_ptr) {
+    int i, j, ret;
+    char **argv = *argv_ptr;
+    char **new_argv;
+    int argc = dcc_argv_len(argv);
+    for (i = 0; argv[i]; i++) {
+        if (str_startswith("-Wp,", argv[i])) {
+            /* First, calculate how many extra arguments we'll need. */
+            int extra_args = count_extra_args(argv[i]);
+            assert(extra_args >= 1);
+
+            new_argv = calloc(argc + extra_args, sizeof(char *));
+            if (!new_argv) {
+                return EXIT_OUT_OF_MEMORY;
+            }
+            for (j = 0; j < i; j++) {
+                new_argv[j] = argv[j];
+            }
+            if ((ret = copy_extra_args(new_argv + i, argv[i],
+                                       extra_args)) != 0) {
+                free(new_argv);
+                return ret;
+            }
+            for (j = i + 1; j <= argc; j++) {
+                new_argv[j + extra_args - 1] = argv[j];
+            }
+            free(argv);
+            *argv_ptr = argv = new_argv;
+        }
+    }
+    return 0;
+}
