@@ -1,23 +1,24 @@
-/* -*- c-file-style: "java"; indent-tabs-mode: nil; fill-column: 78 -*-
- * 
+/* -*- c-file-style: "java"; indent-tabs-mode: nil; tab-width: 4; fill-column: 78 -*-
+ *
  * distcc -- A simple distributed compiler system
  *
- * Copyright (C) 2002, 2003 by Martin Pool <mbp@samba.org>
+ * Copyright (C) 2002, 2003, 2004 by Martin Pool <mbp@samba.org>
+ * Copyright 2007 Google Inc.
  *
  * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
+ * modify it under the terms of the GNU General Public License
+ * as published by the Free Software Foundation; either version 2
+ * of the License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
  *
- * This program is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * General Public License for more details.
- * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
- * USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
+ * USA.
  */
 
 #ifndef _DISTCC_EXITCODE_H
@@ -32,8 +33,11 @@
 /**
  * Common exit codes for both client and server.
  *
- * These need to be in [1,255] so that they can be used as exit()
- * codes.
+ * These need to be in [1,255] so that they can be used as exit() codes.  They
+ * are fairly high so that they're not confused with the real error code from
+ * gcc.
+ *
+ * TODO: Put more of these into the manpage.
  **/
 enum dcc_exitcode {
     EXIT_DISTCC_FAILED            = 100, /**< General failure */
@@ -53,7 +57,8 @@ enum dcc_exitcode {
     EXIT_BUSY                     = 114, /**< In use by another process. */
     EXIT_NO_SUCH_FILE             = 115,
     EXIT_NO_HOSTS                 = 116,
-    EXIT_GONE                     = 117 /**< No longer relevant */
+    EXIT_GONE                     = 117, /**< No longer relevant */
+    EXIT_TIMEOUT                  = 118
 };
 
 
